use std::{io::ErrorKind, time::SystemTime};

use actix_files::NamedFile;
use actix_web::{
    error,
    http::header::{
        Accept, CacheControl, CacheDirective, Charset, ContentDisposition, DispositionParam,
        DispositionType, Expires, ExtendedValue, Header, HeaderValue, HttpDate, TryIntoHeaderValue,
        ACCEPT, CACHE_CONTROL, CONTENT_TYPE, EXPIRES, VARY,
    },
    web, Error, HttpRequest, HttpResponse,
};
use mime::{Mime, APPLICATION_OCTET_STREAM, TEXT_HTML};
use sqlx::postgres::PgPool;
use std::path::Path;
use time::OffsetDateTime;
use tokio::fs;

use crate::{
    config::Config,
    deleter,
    file_info::{self, FileInfo},
    mime_relations, template,
};

const TEXT_VIEW_SIZE_LIMIT: u64 = 512 * 1024; // 512KiB

enum ViewType {
    Raw,
    Download,
    Html,
}

pub async fn download(
    req: HttpRequest,
    db: web::Data<PgPool>,
    config: web::Data<Config>,
) -> Result<HttpResponse, Error> {
    let id = req.match_info().query("id");
    let file_info = file_info::find_by_id(id, &db)
        .await
        .map_err(|db_err| {
            log::error!("could not run select statement {:?}", db_err);
            error::ErrorInternalServerError("could not run select statement")
        })?
        .ok_or_else(|| error::ErrorNotFound("file does not exist or has expired"))?;
    let delete = file_info.delete_on_download;
    let valid_till = file_info.valid_till;
    let path = config.files_dir.join(&file_info.file_id);

    let mut response = match get_view_type(&file_info, &path, &req).await {
        ViewType::Raw => build_file_response(file_info, &path, false, &req),
        ViewType::Download => build_file_response(file_info, &path, true, &req),
        ViewType::Html => build_html_response(file_info, &path, &config, &req).await,
    }?;

    insert_cache_headers(&mut response, valid_till);

    if delete {
        deleter::delete_by_id(&db, id, &config.files_dir)
            .await
            .map_err(|db_err| {
                log::error!("could not delete file {:?}", db_err);
                error::ErrorInternalServerError("could not delete file")
            })?;
    }

    Ok(response)
}

async fn get_view_type(file_info: &FileInfo, file_path: &Path, req: &HttpRequest) -> ViewType {
    if file_info.delete_on_download || req.query_string().contains("dl") {
        return ViewType::Download;
    }
    if req.query_string().contains("raw") {
        return ViewType::Raw;
    }
    if !mime_relations::matches_text(&file_info.content_type) {
        return ViewType::Raw;
    }
    if get_file_size(file_path).await >= TEXT_VIEW_SIZE_LIMIT {
        return ViewType::Raw;
    }
    if let Ok(accept) = Accept::parse(req) {
        for accept_mime in accept.ranked() {
            if accept_mime == TEXT_HTML {
                return ViewType::Html;
            }
            if mime_matches(&accept_mime, &file_info.content_type) {
                break;
            }
        }
    }
    ViewType::Raw
}

fn mime_matches(accept: &Mime, content: &Mime) -> bool {
    let type_matches = accept.type_() == content.type_() || accept.type_() == mime::STAR;
    let subtype_matches = accept.subtype() == content.subtype() || accept.subtype() == mime::STAR;
    type_matches && subtype_matches
}

async fn get_file_size(file_path: &Path) -> u64 {
    fs::metadata(file_path)
        .await
        .map(|metadata| metadata.len())
        .unwrap_or(0)
}

async fn build_html_response(
    file_info: FileInfo,
    path: &Path,
    config: &Config,
    req: &HttpRequest,
) -> Result<HttpResponse, Error> {
    let content = match fs::read_to_string(path).await {
        Ok(content) => content,
        Err(file_err) if file_err.kind() == ErrorKind::InvalidData => {
            // content may not be valid UTF-8, try to return the raw file instead
            return build_file_response(file_info, path, false, req);
        }
        Err(file_err) => {
            log::error!("file could not be read: {:?}", file_err);
            return Err(error::ErrorInternalServerError(
                "this file should be here but could not be found",
            ));
        }
    };
    let html_view =
        template::build_html_view_template(&content, file_info.file_name.as_deref(), req, config);
    Ok(HttpResponse::Ok()
        .content_type(TEXT_HTML.to_string())
        .body(html_view))
}

fn build_file_response(
    file_info: FileInfo,
    path: &Path,
    download: bool,
    req: &HttpRequest,
) -> Result<HttpResponse, Error> {
    let file_name: String = file_info.file_name.unwrap_or_else(|| {
        let file_id = file_info.file_id;
        let extension = mime_relations::get_extension(&file_info.content_type).unwrap_or("txt");
        format!("{file_id}.{extension}")
    });
    let content_disposition = ContentDisposition {
        disposition: if download {
            DispositionType::Attachment
        } else {
            DispositionType::Inline
        },
        parameters: get_disposition_params(file_name),
    };
    let file = NamedFile::open(path)
        .map_err(|file_err| {
            log::error!("file could not be read: {:?}", file_err);
            error::ErrorInternalServerError("this file should be here but could not be found")
        })?
        .set_content_type(file_info.content_type)
        .set_content_disposition(content_disposition);

    let mut response = file.into_response(req);
    append_security_headers(&mut response, req);
    Ok(response)
}

fn get_disposition_params(filename: String) -> Vec<DispositionParam> {
    if !filename.is_ascii() {
        vec![
            DispositionParam::Filename(filename.clone()),
            DispositionParam::FilenameExt(ExtendedValue {
                charset: Charset::Ext(String::from("UTF-8")),
                language_tag: None,
                value: filename.into_bytes(),
            }),
        ]
    } else {
        vec![DispositionParam::Filename(filename)]
    }
}

const ALLOWED_CONTEXTS: [&str; 6] = ["audio", "document", "empty", "font", "image", "video"];

fn append_security_headers(response: &mut HttpResponse, req: &HttpRequest) {
    // if the browser is trying to fetch this resource in a secure context pretend the response is
    // just binary data so it won't be executed
    let sec_fetch_dest = req
        .headers()
        .get("sec-fetch-dest")
        .map(|v| v.to_str().unwrap_or("unknown"));
    if sec_fetch_dest.is_some_and(|sec_fetch_dest| !ALLOWED_CONTEXTS.contains(&sec_fetch_dest)) {
        response.headers_mut().insert(
            CONTENT_TYPE,
            HeaderValue::from_str(APPLICATION_OCTET_STREAM.as_ref())
                .expect("mime type can be encoded to header value"),
        );
    }
    // the response varies based on these request headers
    response
        .headers_mut()
        .append(VARY, HeaderValue::from_static("sec-fetch-dest"));
}

fn insert_cache_headers(response: &mut HttpResponse, valid_till: OffsetDateTime) {
    if response.status().is_success() {
        let valid_duration = valid_till - OffsetDateTime::now_utc();
        let valid_cache_seconds =
            valid_duration.whole_seconds().clamp(0, i64::from(u32::MAX)) as u32;
        response.headers_mut().insert(
            CACHE_CONTROL,
            CacheControl(vec![
                CacheDirective::Public,
                CacheDirective::MustRevalidate,
                CacheDirective::MaxAge(valid_cache_seconds),
                CacheDirective::NoTransform,
                CacheDirective::Extension("immutable".to_owned(), None),
            ])
            .try_into_value()
            .unwrap(),
        );
        response.headers_mut().insert(
            EXPIRES,
            Expires(HttpDate::from(
                SystemTime::now() + std::time::Duration::from_secs(valid_cache_seconds.into()),
            ))
            .try_into_value()
            .unwrap(),
        );
    }
    response
        .headers_mut()
        .append(VARY, HeaderValue::from_name(ACCEPT));
}
