use mime::{Mime, APPLICATION_OCTET_STREAM};
use sqlx::{postgres::PgRow, FromRow, Row};
use time::OffsetDateTime;

use crate::multipart::UploadConfig;

pub struct FileInfo {
    pub file_id: String,
    pub file_name: Option<String>,
    pub valid_till: OffsetDateTime,
    pub content_type: Mime,
    pub delete_on_download: bool,
}

impl FileInfo {
    pub fn new(file_id: String, upload_config: UploadConfig) -> Self {
        Self {
            file_id,
            file_name: upload_config.original_name,
            valid_till: upload_config.valid_till,
            content_type: upload_config.content_type,
            delete_on_download: upload_config.delete_on_download,
        }
    }
}

pub async fn find_by_id(
    id: &str,
    db: &sqlx::Pool<sqlx::Postgres>,
) -> Result<Option<FileInfo>, sqlx::Error> {
    sqlx::query_as(
        "SELECT file_id, file_name, valid_till, content_type, delete_on_download from files WHERE file_id = $1",
    )
    .bind(id)
    .fetch_optional(db)
    .await
}

pub async fn create(file_info: &FileInfo, db: &sqlx::Pool<sqlx::Postgres>) -> Result<(), sqlx::Error> {
    sqlx::query(
        "INSERT INTO Files (file_id, file_name, content_type, valid_till, delete_on_download) \
        VALUES ($1, $2, $3, $4, $5)",
    )
    .bind(&file_info.file_id)
    .bind(&file_info.file_name)
    .bind(file_info.content_type.to_string())
    .bind(file_info.valid_till)
    .bind(file_info.delete_on_download)
    .execute(db)
    .await
    .map(|_| ())
}

impl FromRow<'_, PgRow> for FileInfo {
    fn from_row(row: &'_ PgRow) -> Result<Self, sqlx::Error> {
        Ok(Self {
            file_id: row.try_get("file_id")?,
            file_name: row.try_get("file_name")?,
            valid_till: row.try_get("valid_till")?,
            content_type: row
                .try_get_raw("content_type")?
                .as_str()
                .map_err(sqlx::Error::Decode)?
                .parse()
                .unwrap_or(APPLICATION_OCTET_STREAM),
            delete_on_download: row.try_get("delete_on_download")?,
        })
    }
}
